from .config import Config
from .template import Template


def test(config: Config, template: Template) -> None:
    variable_names = set(variable.name for variable in config.variables)
    for file_template in template.file_templates:
        for variable_name in file_template.variables:
            assert (
                variable_name in variable_names
            ), f"Undefined variable '{variable_name}' in {file_template.template_path}"



import click
import pathlib
import subprocess
import typing
import yaml

from click.decorators import FC
from collections.abc import Callable
from dataclasses import dataclass, field
from dataclasses_json import config, DataClassJsonMixin
from dataclasses_json.core import Json


@dataclass
class DefaultConfig(DataClassJsonMixin):
    shell: str

    @staticmethod
    def encoder(
        default_spec: typing.Union[None, str, "DefaultConfig"],
    ) -> Json:
        if default_spec is None or isinstance(default_spec, str):
            return default_spec
        elif isinstance(default_spec, dict):
            return DefaultConfig.to_dict(default_spec)
        else:
            raise TypeError(default_spec)

    @staticmethod
    def decoder(
        kvs: Json,
    ) -> typing.Union[None, str, "DefaultConfig"]:
        if kvs is None or isinstance(kvs, str):
            return kvs
        elif isinstance(kvs, dict):
            return DefaultConfig.from_dict(kvs)
        else:
            raise TypeError(kvs)


@dataclass
class VariableConfig(DataClassJsonMixin):
    name: str
    default: typing.Union[None, str, DefaultConfig] = field(
        default=None,
        metadata=config(decoder=DefaultConfig.decoder, encoder=DefaultConfig.encoder),
    )

    @property
    def has_default(self) -> bool:
        return self.default is not None

    def get_prompt(self) -> str:
        return f"Enter {' '.join(self.name.split('_'))}"

    def get_default(self) -> typing.Optional[str]:
        if isinstance(self.default, DefaultConfig):
            try:
                self.default = subprocess.getoutput(self.default.shell)
            except subprocess.CalledProcessError:
                return None
        return self.default

    def option(self) -> click.Option:
        if self.has_default:
            return click.Option(
                param_decls=[f"--{self.name}"],
                prompt=self.get_prompt(),
                default=self.get_default(),
            )
        else:
            return click.Option(
                param_decls=[f"--{self.name}"],
                prompt=self.get_prompt(),
            )


@dataclass
class Config:
    name: str
    variables: list[VariableConfig] = field(default_factory=list)

    @staticmethod
    def load(path: typing.Union[str, pathlib.Path]) -> "Config":
        if isinstance(path, str):
            path = pathlib.Path(path)
        contents = yaml.safe_load(path.read_text())
        variables = []
        name = contents["name"]
        for variable_dict in contents["variables"]:
            variables.append(VariableConfig.from_dict(variable_dict))
        return Config(name=name, variables=variables)

    def option_parser(self, ctx: click.Context) -> click.OptionParser:
        """
        Create a click option parser for a given variable spec.
        """
        option_parser = click.OptionParser(ctx)
        for spec in self.variables:
            spec.option().add_to_parser(option_parser, ctx)
        return option_parser

    def command(self, **kwargs) -> click.Command:
        """
        Create a click option parser for a given variable spec.
        """
        return click.Command(
            name=self.name,
            params=[spec.option() for spec in self.variables],
            **kwargs,
        )


import os
from pathlib import Path
import click
import itertools
import sys
import tempfile
import wget
import zipfile

from metaproject.test import test

from .config import Config
from .template import Template


@click.group(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.pass_context
def cli(ctx: click.Context) -> None:
    pass


@cli.command(
    name="init",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.pass_context
@click.argument("repo", type=str)
def init(
    ctx: click.Context,
    repo: str,
) -> None:
    with tempfile.TemporaryDirectory(prefix="metaproject") as temp_dir:

        # Download repository zipfile from GitHub:
        repo_zip_file = wget.download(
            url=f"https://github.com/{repo}/archive/refs/heads/dev.zip",
            out=f"{temp_dir}/{repo.replace('/','_')}.zip",
            bar=lambda current, total, width: (),
        )

        # Extract repository zipfile from GitHub:
        repo_zip_obj = zipfile.ZipFile(repo_zip_file)
        repo_zip_obj.extractall(path=temp_dir)

        # Parse metaproject configuration:
        template_dir: Path
        config: Config
        for config_path in Path(temp_dir).glob(f"*/metaproject-config.yaml"):
            template_dir = config_path.parent
            config = Config.load(config_path)

        # Parse metaproject template:
        template: Template = Template(template_dir)

        # Verify metaproject:
        test(config, template)

        command = config.command(callback=template.init)
        command.parse_args(ctx, ctx.args)
        command.invoke(ctx)


if __name__ == "__main__":
    cli()


# confirmation_dialog = "\n".join(
#     [
#         "Initialize project with the following options:",
#         "\n".join([f"{key}: {val}" for key, val in kwargs.items()]),
#         "Ok?",
#     ]
# )
# click.confirm(text=confirmation_dialog)
# output_dir = pathlib.Path(__file__).parent.parent
# template_dir = pathlib.Path(__file__).parent / ".template"
# env = jinja2.Environment()
# for template_path in template_dir.glob("**/*"):
#     output_path_template: str = str(
#         template_path.relative_to(template_dir)
#     ).replace(".jinja2", "")
#     output_path = env.from_string(source=output_path_template).render(**kwargs)
#     print(f"{output_path_template} -> {output_path}")
#     output_path = project_dir / output_path
#     if template_path.is_dir():
#         output_path.mkdir(parents=True, exist_ok=True)
#     else:
#         output_path_contents = env.get_template(output_path_template).render(**kwargs)
#         output_path.write_text(output_path_contents)



import click
import jinja2
import jinja2.meta
import pathlib
import typing

PathTemplate = jinja2.Template

ContentsTemplate = jinja2.Template


class FileTemplate:
    template_path: pathlib.Path
    path_template: PathTemplate
    contents_template: ContentsTemplate
    variables: set[str]

    def __init__(
        self,
        template_path: pathlib.Path,
        *,
        template_dir: typing.Optional[pathlib.Path] = None,
        env: typing.Optional[jinja2.Environment] = None,
    ):
        # Save the template path
        self.template_path = template_path

        # Use default environment
        env = env or jinja2.Environment()

        # Create the path_template
        if template_dir:
            path_template_str = str(template_path.relative_to(template_dir))
        else:
            path_template_str = str(template_path)
        self.path_template = env.from_string(source=path_template_str)

        # Create the contents_template
        contents_template_str = template_path.read_text()
        self.contents_template = env.from_string(source=contents_template_str)

        # Store variables used in path_template and contents_template
        self.variables = set(
            (
                *jinja2.meta.find_undeclared_variables(env.parse(path_template_str)),
                *jinja2.meta.find_undeclared_variables(
                    env.parse(contents_template_str)
                ),
            )
        )

    def output_path(self, context: typing.Any) -> pathlib.Path:
        return pathlib.Path(self.path_template.render(context))

    def output_file_contents(self, context: typing.Any) -> str:
        return self.contents_template.render(context)


class Template:
    file_templates: typing.Sequence[FileTemplate]

    def __init__(
        self,
        template_dir: pathlib.Path,
        *,
        env: typing.Optional[jinja2.Environment] = None,
    ):
        self.file_templates = []
        for template_path in template_dir.glob("**/*"):
            if (
                template_path.is_file()
                and template_path.name != "metaproject-config.yaml"
            ):
                self.file_templates.append(
                    FileTemplate(template_path, template_dir=template_dir, env=env)
                )

    def init(self, **context):
        # Render file paths:
        output_paths = []
        for file_template in self.file_templates:
            output_paths.append(str(file_template.output_path(context)))

        # Ask for confirmation:
        click.confirm(
            "\n".join(
                [
                    "Create the following files?",
                    *output_paths,
                ]
            )
        )

        for file_template in self.file_templates:
            # Render file path:
            output_path = file_template.output_path(context)
            output_path.parent.mkdir(parents=True, exist_ok=True)
            # Render file contents:
            output_contents = file_template.output_file_contents(context)
            output_path.write_text(output_contents)
